{-# LANGUAGE DeriveAnyClass #-}
{-# LANGUAGE ScopedTypeVariables #-}

{-|
Module      : OpenAlex.Types
Description : Textmining Collaborative Platform
Copyright   : (c) CNRS, 2023-Present
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

@Gargantext@: search, map, share
-}

module OpenAlex.Types where

import Control.Monad.Fail (fail)
import Data.Aeson
import Data.Csv qualified as Csv
import Data.Scientific (floatingOrInteger)
import Data.Text qualified as T
import Data.Time (UTCTime(..))
import Data.Time.Calendar (Day, toGregorian)
import Data.Time.Format qualified as DTF
import Data.Vector qualified as V
import OpenAlex.Utils (reconstructAbstract)
import Protolude hiding (Location, Meta)
import Protolude.Base (Show(..))

-- API request types
type Cursor = Text
type Page = Int
type PerPage = Int

type Filter = Text
type Search = Text

-- API response types


type Count = Int
type DOI = Text
data ExternalID = ExtIDUrl URL | ExtIDUrls [URL] | ExtIDInt Int
  deriving (Generic, Show)
instance FromJSON ExternalID where
  parseJSON (String v) = pure $ ExtIDUrl v
  parseJSON (Number n) =
    case (floatingOrInteger n :: Either Double Integer) of
      Left _ -> fail "Floating number not supported as external id"
      Right i -> pure $ ExtIDInt $ fromIntegral i
  parseJSON a@(Array _a) =do
    ids <- parseJSONList a
    pure $ ExtIDUrls ids
  parseJSON _ = fail "Don't know how to handle this external id"
type ISSN = Text
type ISSNL = Text
type Language = Text  -- TODO: https://doc.wikimedia.org/mediawiki-core/master/php/Names_8php_source.html
type Level = Int
-- |https://docs.openalex.org/api-entities/works/work-object#oa_status
data OAStatus = OAGold | OAGreen | OAHybrid | OABronze | OAClosed | OADiamond
  deriving (Generic, Show)
instance FromJSON OAStatus where
  parseJSON (String "gold") = pure OAGold
  parseJSON (String "green") = pure OAGreen
  parseJSON (String "hybrid") = pure OAHybrid
  parseJSON (String "bronze") = pure OABronze
  parseJSON (String "closed") = pure OAClosed
  parseJSON (String "diamond") = pure OADiamond
  parseJSON (String str) = fail $ "Don't know how to parse this oa status: " <> Protolude.show str
  parseJSON _ = fail "OAStatus should be a string"
type OpenAlexID = Text
type URL = Text
type Year = Int

-- newtype CreatedDate = CreatedDate Day
--   deriving (Generic, Show)
-- instance FromJSON CreatedDate
-- newtype UpdatedDate = UpdatedDate Day
--   deriving (Generic, Show)
-- instance FromJSON UpdatedDate

parseTimeE :: (MonadFail m, DTF.ParseTime t) => Text -> Text -> m t
parseTimeE fmt s = case (DTF.parseTimeM True DTF.defaultTimeLocale (T.unpack fmt) (T.unpack s)) of
  Nothing -> fail $ "Cannot parse date with format " <> T.unpack fmt
  Just  p -> pure p

-- | Convert to ISO text format
showDate :: Date -> Text
showDate (DDay day) = T.pack $ DTF.formatTime DTF.defaultTimeLocale "%Y-%m-%dT00:00:00" day
showDate (DUTCTime t) = T.pack $ DTF.formatTime DTF.defaultTimeLocale "%Y-%m-%dT%H:%M:%S" t

data Date = DDay Day | DUTCTime UTCTime
  deriving (Generic)
instance Show Date where
  show = T.unpack . showDate
instance FromJSON Date where
  parseJSON = withText "Date" $ \s ->
    (DDay <$> parseTimeE "%F" s) <|>
    (DUTCTime <$> parseTimeE "%Y-%m-%dT%H:%M:%S%Q" s)

-- | Return the 'day' part of `Date`
dateDay :: Date -> Int
dateDay (DDay d) = day
  where
    (_year, _month, day) = toGregorian d
dateDay (DUTCTime t) = dateDay (DDay (utctDay t))

-- | Return the 'month' part of `Date`
dateMonth :: Date -> Int
dateMonth (DDay d) = month
  where
    (_year, month, _day) = toGregorian d
dateMonth (DUTCTime t) = dateMonth (DDay (utctDay t))

-- | Return the 'year' part of `Date`
dateYear :: Date -> Integer
dateYear (DDay d) = year
  where
    (year, _month, _day) = toGregorian d
dateYear (DUTCTime t) = dateYear (DDay (utctDay t))

type CreatedDate = Date
type UpdatedDate = Date

-- https://docs.openalex.org/api-entities/concepts/concept-object#ids
data ExternalDB = MAG | OpenAlex | UMLS_Cui | UMLS_Aui | Wikidata | Wikipedia
  deriving (Show, Eq, Ord, Generic, Hashable, FromJSONKey)
instance FromJSON ExternalDB where
  parseJSON (String "mag") = pure MAG
  parseJSON (String "openalex") = pure OpenAlex
  parseJSON (String "umls_cui") = pure UMLS_Cui
  parseJSON (String "umls_aui") = pure UMLS_Aui
  parseJSON (String "wikidata") = pure Wikidata
  parseJSON (String "wikipedia") = pure Wikipedia
  parseJSON _ = fail "Don't know how to handle this external db"

data Meta = Meta
  { count               :: Count
  , db_response_time_ms :: Int
  , page                :: Maybe Page
  , per_page            :: PerPage
  -- | https://docs.openalex.org/how-to-use-the-api/get-lists-of-entities/paging
  , next_cursor         :: Maybe Cursor
  } deriving (Generic, Show)
instance FromJSON Meta

data ListOf a = ListOf
  { meta     :: Meta
  , results  :: [a]
  , group_by :: [Int] -- Not implemented
  } deriving (Generic, Show)
instance FromJSON a => FromJSON (ListOf a)

-- | https://docs.openalex.org/api-entities/concepts/concept-object
data Concept = Concept
  { ancestors           :: [Ancestor]
  , cited_by_count      :: Count
  , counts_by_year      :: [CountByYear]
  , created_date        :: CreatedDate
  , description         :: Text
  , display_name        :: Text
  , id                  :: OpenAlexID
  , ids                 :: Map Text ExternalID  -- TODO ExternalDB
  , image_thumbnail_url :: URL
  , image_url           :: URL
  , international       :: Map Language Text
  , level               :: Level
  , related_concepts    :: [DehydratedConcept]
  , summary_stats       :: SummaryStats
  , updated_date        :: UpdatedDate
  , wikidata            :: URL
  , works_api_url       :: URL
  , works_count         :: Count
  } deriving (Generic, Show)
instance FromJSON Concept where
  parseJSON (Object v) = do
    ancestors <- v .: "ancestors"
    cited_by_count <- v .: "cited_by_count"
    counts_by_year <- v .: "counts_by_year"
    created_date <- v .: "created_date"
    description <- v .: "description"
    display_name <- v .: "display_name"
    id <- v .: "id"
    ids <- v .: "ids"
    image_thumbnail_url <- v .: "image_thumbnail_url"
    image_url <- v .: "image_url"
    international' <- v .: "international"
    international <- international' .: "display_name"
    level <- v .: "level"
    related_concepts <- v .: "related_concepts"
    summary_stats <- v .: "summary_stats"
    updated_date <- v .: "updated_date"
    wikidata <- v .: "wikidata"
    works_api_url <- v .: "works_api_url"
    works_count <- v .: "works_count"
    pure $ Concept { .. }
  parseJSON _ = fail "Cannot parse Concept as a non-object"

-- | https://docs.openalex.org/api-entities/concepts/concept-object#the-dehydratedconcept-object
data DehydratedConcept = DehydratedConcept
  { display_name :: Text
  , id           :: OpenAlexID
  , level        :: Level
  , wikidata     :: Maybe URL
  } deriving (Generic, Show)
instance FromJSON DehydratedConcept

data Ancestor = Ancestor
  { id           :: URL
  , wikidata     :: URL
  , display_name :: Text
  , level        :: Level
  } deriving (Generic, Show)
instance FromJSON Ancestor

data CountByYear = CountByYear
  { year           :: Year
  , works_count    :: Maybe Count
  , cited_by_count :: Count
  } deriving (Generic, Show, FromJSON)


data SummaryStats = SummaryStats
  { two_year_mean_citedness :: Double
  , h_index :: Int
  , i10_index :: Int
  } deriving (Generic, Show)
instance FromJSON SummaryStats where
  parseJSON (Object v) =
    SummaryStats <$> v .: "2yr_mean_citedness"
                 <*> v .: "h_index"
                 <*> v .: "i10_index"
  parseJSON _ = fail "Don't know how to parse this as SummaryStats"


-- | https://docs.openalex.org/api-entities/works/work-object
data Work = Work
  { abstract_inverted_index       :: Maybe (Map Text [Int])  -- TODO
  , abstract_reconstructed        :: Text
  , authorships                   :: [Authorship]
  , apc_list                      :: Maybe APCList
  , apc_paid                      :: Maybe APCPaid
  , best_oa_location              :: Maybe Location
  , biblio                        :: Biblio
  , cited_by_api_url              :: Text
  , cited_by_count                :: Count
  , concepts                      :: [DehydratedConcept]
  , corresponding_author_ids      :: [OpenAlexID]
  , corresponding_institution_ids :: [OpenAlexID]
  , counts_by_year                :: [CountByYear]
  , created_date                  :: CreatedDate
  , display_name                  :: Maybe Text
  , doi                           :: Maybe DOI
  , grants                        :: [Grant]
  , id                            :: OpenAlexID
  , ids                           :: Map Text ExternalID  -- TODO ExternalDB
  , is_paratext                   :: Bool
  , is_retracted                  :: Bool
  , language                      :: Maybe Text
  , locations                     :: [Location]
  , locations_count               :: Count
  , mesh                          :: [MeSH]
  , ngrams_url                    :: Maybe URL
  , open_access                   :: OpenAccess
  , primary_location              :: Maybe Location
  , publication_date              :: CreatedDate
  , publication_year              :: Year
  , referenced_works              :: [OpenAlexID]
  , related_works                 :: [OpenAlexID]
  , title                         :: Maybe Text
  , type_                         :: Text
  , updated_date                  :: UpdatedDate
  , is_oa                         :: Maybe Bool
  , license                       :: Maybe Text
  , url                           :: Maybe URL
  , version                       :: Maybe Text
  } deriving (Generic, Show)
instance FromJSON Work where
  parseJSON = withObject "Work" $ \v -> do
    abstract_inverted_index <- v .: "abstract_inverted_index"
    let abstract_reconstructed = reconstructAbstract abstract_inverted_index
    authorships <- v .: "authorships"
    apc_list <- v .: "apc_list"
    apc_paid <- v .: "apc_paid"
    best_oa_location <- v .: "best_oa_location"
    biblio <- v .: "biblio"
    cited_by_api_url <- v .: "cited_by_api_url"
    cited_by_count <- v .: "cited_by_count"
    concepts <- v .: "concepts"
    corresponding_author_ids <- v .: "corresponding_author_ids"
    corresponding_institution_ids <- v .: "corresponding_institution_ids"
    counts_by_year <- v .: "counts_by_year"
    created_date <- v .: "created_date"
    display_name <- v .:? "display_name"
    doi <- v .:? "doi"
    grants <- v .: "grants"
    id <- v .: "id"
    ids <- v .: "ids"
    is_paratext <- v .: "is_paratext"
    is_retracted <- v .: "is_retracted"
    language <- v .:? "language"
    locations <- v .: "locations"
    locations_count <- v .: "locations_count"
    mesh <- v .: "mesh"
    ngrams_url <- v .:? "ngrams_url"
    open_access <- v .: "open_access"
    primary_location <- v .:? "primary_location"
    publication_date <- v .: "publication_date"
    publication_year <- v .: "publication_year"
    referenced_works <- v .: "referenced_works"
    related_works <- v .: "related_works"
    title <- v .:? "title"
    type_ <- v .: "type"
    updated_date <- v .: "updated_date"
    is_oa <- v .:? "is_oa"
    license <- v .:? "license"
    url <- v .:? "url"
    version <- v .:? "version"
    pure $ Work { .. }
-- | Publication Day,Publication Month,Publication Year,Authors,Title,Source,Abstract
instance Csv.DefaultOrdered Work where
  headerOrder _ = V.fromList [ "Publication Day"
                             , "Publication Month"
                             , "Publication Year"
                             , "Authors"
                             , "Title"
                             , "Source"
                             , "Abstract" ]
instance Csv.ToNamedRecord Work where
  toNamedRecord (Work { .. }) = Csv.namedRecord [
      Csv.namedField "Publication Day" (dateDay $ publication_date)
    , Csv.namedField "Publication Month" (dateMonth $ publication_date)
    , Csv.namedField "Publication Year" (dateYear $ publication_date)
    , Csv.namedField "Authors" (T.intercalate ", " authorList)
    , Csv.namedField "Title" (fromMaybe "" title)
    , Csv.namedField "Source" source
    , Csv.namedField "Abstract" abstract_reconstructed
    ]
    where
      authorList :: [Text]
      authorList = catMaybes (authorshipAuthorName <$> authorships)
      source :: Text
      source = case primary_location of
        Nothing -> ""
        Just (Location { source = s }) ->
          case s of
            Nothing -> ""
            Just (DehydratedSource { display_name = dn }) -> dn

data APCList = APCList
  { value      :: Maybe Int
  , currency   :: Text
  , provenance :: Maybe Text
  , value_usd  :: Maybe Int
  } deriving (Generic, Show, FromJSON)

data APCPaid = APCPaid
  { value      :: Int
  , currency   :: Text
  , provenance :: Maybe Text
  , value_usd  :: Maybe Int
  } deriving (Generic, Show, FromJSON)

-- | https://docs.openalex.org/api-entities/works/work-object/authorship-object
data Authorship = Authorship
  { author                 :: DehydratedAuthor
  , author_position        :: Text
  , institutions           :: [DehydratedInstitution]
  , is_corresponding       :: Maybe Bool
  , raw_affiliation_string :: Maybe Text
  } deriving (Generic, Show, FromJSON)
authorshipAuthorName :: Authorship -> Maybe Text
authorshipAuthorName (Authorship { author = DehydratedAuthor { display_name } }) = display_name

data Biblio = Biblio
  { volume     :: Maybe Text
  , issue      :: Maybe Text
  , first_page :: Maybe Text
  , last_page  :: Maybe Text
  } deriving (Generic, Show, FromJSON)

data DehydratedAuthor = DehydratedAuthor
  { id           :: OpenAlexID
  , display_name :: Maybe Text
  , orcid        :: Maybe URL
  } deriving (Generic, Show, FromJSON)

data DehydratedInstitution = DehydratedInstitution
  { id           :: Maybe OpenAlexID
  , display_name :: Text
  , ror          :: Maybe Text
  , country_code :: Maybe Text
  , type_        :: Maybe Text
  } deriving (Generic, Show)
instance FromJSON DehydratedInstitution where
  parseJSON (Object v) = do
    id <- v .:? "id"
    display_name <- v .: "display_name"
    ror <- v .:? "ror"
    country_code <- v .:? "country_code"
    type_ <- v .:? "type"
    pure $ DehydratedInstitution { .. }
  parseJSON _ = fail "Don't know how to parse a dehydrated institution from a non-object"

data Grant = Grant
  { funder              :: OpenAlexID
  , funder_display_name :: Text
  , award_id            :: Maybe Text
  } deriving (Generic, Show, FromJSON)

data Location = Location
  { is_oa            :: Bool
  , landing_page_url :: Maybe URL
  , license          :: Maybe Text
  , source           :: Maybe DehydratedSource
  , pdf_url          :: Maybe URL
  , version          :: Maybe Text
  } deriving (Generic, Show, FromJSON)

-- | PubMed only, https://docs.openalex.org/api-entities/works/work-object#mesh
data MeSH = MeSH
  { descriptor_ui   :: Text
  , descriptor_name :: Text
  , qualifier_ui    :: Text
  , qualifier_name  :: Maybe Text
  , is_major_topic  :: Bool
  } deriving (Generic, Show, FromJSON)

-- | https://docs.openalex.org/api-entities/works/work-object#the-openaccess-object
data OpenAccess = OpenAccess
  { any_repository_has_fulltext :: Bool
  , is_oa                       :: Bool
  , oa_status                   :: OAStatus
  , oa_url                      :: Maybe URL
  } deriving (Generic, Show, FromJSON)

-- | https://docs.openalex.org/api-entities/sources/source-object#the-dehydratedsource-object
data DehydratedSource = DehydratedSource
  { display_name :: Text
  , host_organization :: Maybe Text
  , host_organization_lineage :: [OpenAlexID]
  , host_organization_name :: Maybe Text
  , id :: OpenAlexID
  , is_in_doaj :: Bool
  , issn :: [ISSN]
  , issn_l :: Maybe ISSNL
  , type_ ::Text
  } deriving (Generic, Show)
instance FromJSON DehydratedSource where
  parseJSON = withObject "DehydratedSource" $ \v -> do
    display_name <- v .: "display_name"
    host_organization <- v .:? "host_organization"
    host_organization_lineage' <- v .:? "host_organization_lineage"
    let host_organization_lineage = fromMaybe [] host_organization_lineage'
    host_organization_name <- v .:? "host_organization_name"
    id <- v .: "id"
    is_in_doaj <- v .: "is_in_doaj"
    issn' <- v .:? "issn"
    let issn = fromMaybe [] issn'
    issn_l <- v .:? "issn_l"
    type_ <- v .: "type"
    pure $ DehydratedSource { .. }
